#include "config.h"
#include "config_funcs.h"
#include "drive_funcs.h"
#include "server.h"
#include "drive.h"
#include "includes.h"
#include "sizes.h"
#include "petscii_funcs.h"
#include "debug_funcs.h"
#include "string_funcs.h"

static int config_get_line(FILE* config, char* line) {
    size_t i;
    while (fgets(line, 1023, config) != NULL) {
        i = strspn(line, " \t\n\v\r");
        //sort out comments
        if(line[i] == '#') continue;
        //sort out empty lines
        else if(i == strlen(line)) continue;
        else return 1;
    }
    return 0;
}

int config_read(server* s, const char* config_file) {
    FILE* config = NULL;
    int res;
    unsigned int umask;
    struct passwd* uid;
    struct group* gid;
    int section = SECT_SERVER;
    struct stat buf;
    int par;
    int i;
    char* key;
    char* val;
    char* line;
    char* path;
    int err = 0;
    drive* curr_drive = NULL;

    key = calloc(1024, sizeof(char));
    val = calloc(1024, sizeof(char));
    line = calloc(1024, sizeof(char));
    path = calloc(1024, sizeof(char));

    if((config=fopen(config_file,"rb"))==NULL) {
        fatal_message("Could not open config file %s\n",config_file);
        err = 1;
    }
    //XXX alloc clienst array to have dynamic amount of drives
    if(config) {
        while(config_get_line(config, line)) {
            key[0] = '\0';
            val[0] = '\0';
            res = sscanf(line, "%1023s %1023s", key, val);
            if(key[0] == '[') {
                //drive section
                if(strncasecmp(key, "[drive",7) == 0) {
                    //XXX warning if no drivename given, or ask for unique drives sections?
                    if(val[strlen(val) - 1] == ']') { val[strlen(val) - 1] = '\0'; }
                    curr_drive = drive_create(s);
                    //XXX check of drive creation fails!

                    dyn_strcpy(&curr_drive->name, val);
                    debug_message("Configuring drive: %s\n", curr_drive->name);
                    section = SECT_DRIVE;
                }
                else if(strncasecmp(key, "[server", 7) == 0) {
                    section = SECT_SERVER;
                }
            }
            //values from section
            else {
                switch(section) {
                    case SECT_SERVER:
                        debug_message("Configuring server: %s %s\n",key,val);
                        if(strcasecmp(key,"port") == 0) {
                            s->port = atoi(val);
                            info_message("listening and talking on port %d %s\n", s->port, val);
                        }
                        if(strcasecmp(key,"debug") == 0) {
                            if(strcasecmp(val, "on") == 0 || atoi(val) > 0) {
                                debug_level = 1;
                            }
                            else {
                                debug_level = 0;
                            }
                        }
                        if(strcasecmp(key, "inject") == 0) {
                            if(strcasecmp(val, "on") == 0 || atoi(val) > 0) {
                                s->inject_mode = 1;
                            }
                            else s->inject_mode = 0;
                        }
                    break;
                    case SECT_DRIVE:
                        //XXX Fatal error on missing val! no key is no problem, then we have also no val! == empty line
                        debug_message("Configuring drive %s: %s='%s'\n", curr_drive->name, key, val);
                        if(strcasecmp(key, "uid") == 0) {
                            uid = getpwnam(val);
                            if(uid == NULL) {
                                fatal_message("user '%s' does not exist.\n", val);
                                err = 1;
                            }
                            //copy from struct
                            else {
                                curr_drive->uid = uid->pw_uid;
                            }
                        }
                        else if(strcasecmp(key, "gid") == 0) {
                            gid = getgrnam(val);
                            if(gid == NULL) {
                                fatal_message("group '%s' does not exist.\n", val);
                                err = 1;
                            }
                            //copy from struct
                            else {
                                curr_drive->gid = gid->gr_gid;
                            }
                        }
                        else if(strcasecmp(key, "umask") == 0) {
                            umask = 0;
                            //do some sanity checks
                            if(strlen(val) < 3 || val[0] < 0x30 || val[0] > 0x37  ||  val[1] < 0x30 || val[1] > 0x37  ||  val[2] < 0x30 || val[2] > 0x37) {
                                fatal_message("illegal value for umask '%s'\n", val);
                                err = 1;
                            }
                            //read in as octal
                            sscanf(val, "%o", &umask);
                            curr_drive->umask = umask;
                        }
                        else if(strcasecmp(key, "ip") == 0) {
                            //XXX do checks on len and correct numbers!
                            strcpy(curr_drive->ip, val);
                        }
                        else if(strcasecmp(key, "partition") == 0) {
                            sscanf(val, "%d:%1023s", &par, path);
                            //check partition number
                            if(par < 0 || par >= MAX_PARTITIONS) {
                                fatal_message("illegal partition %d in '%s'\n", par, val);
                                err = 1;
                            }
                            //check for a path
                            i = strlen(path);
                            if(!i) {
                                fatal_message("no path given for partition '%s'\n", val);
                                err = 1;
                            }

                            //remove trailing slashes
                            while(i > 1 && path[i] == '/') { path[i] = '\0'; i--; }

                            //no leading / on path?! -> relative
                            if(path[0] != '/') {
                                fatal_message("partition path '%s' must be absolute\n", val);
                                err = 1;
                            }

                            //file/dir exists?!
                            stat(path, &buf);

                            //all sane now, copy path
                            //ascii2petscii((uint8_t*)path);
                            dyn_strcpy(&curr_drive->chroot[par], path);
                            //printf("%s\n",path);
                            debug_message("Configuring drive %s: chroot on partition %d ='%s'\n", curr_drive->name, par, curr_drive->chroot[par]);
                        }
                    break;
                }
            }
        }
    }
    free(val);
    free(key);
    free(line);
    free(path);
    fclose(config);
    return err;
}
