/* file.c
 * handles all file access and operations, either buffered or directly */

//rename
//format
//delete
//get_file_type (d64,dir,prg.seq,usw)

#include "file.h"
#include "container.h"
#include "filetypes.h"
#include "fs_d64_funcs.h"
#include "petscii_funcs.h"
#include "debug_funcs.h"
#include "file_funcs.h"
#include "fs_sys_funcs.h"
#include "path_funcs.h"
#include "container_funcs.h"
#include "string_funcs.h"
#include "buffer_funcs.h"
#include "dirent_funcs.h"
#include "buffer.h"

file64* file_create() {
    file64* f = calloc(1, sizeof(file64));
    f->f_target = NULL;
    f->f_source = NULL;
    f->f_command = NULL;
    f->f_header = NULL;
    f->f_glob = NULL;
    f->f_filespec = NULL;
    f->f_path = NULL;
    f->f_id = NULL;
    f->petscii_name = NULL;
    f->ufs_name = NULL;
    f->path = NULL;
    f->cnt = container_create();
    f->buffer = buffer_create();
    return f;
}

void file_destruct(file64* f) {
    if(!f) return;
    if(f->f_source) free(f->f_source);
    if(f->f_target) free(f->f_target);
    if(f->f_command) free(f->f_command);
    if(f->f_header) free(f->f_header);
    if(f->f_glob) free(f->f_glob);
    if(f->f_filespec) free(f->f_filespec);
    if(f->f_path) free(f->f_path);
    if(f->f_id) free(f->f_id);
    if(f->petscii_name) free(f->petscii_name);
    if(f->ufs_name) free(f->ufs_name);
    if(f->path) free(f->path);
    container_destruct(f->cnt);
    f->cnt = NULL;
    buffer_destruct(f->buffer);
    f->buffer = NULL;
    free(f);
}

void file_init(file64* f) {
    f->buffered = 0;
    f->open = 0;
    f->err = 0;
    f->err_track  = 0;
    f->err_sector = 0;
    f->blocks = 0;
    f->track = 0;
    f->sector = 0;
    f->detrack = 0;
    f->desector = 0;
    f->desectpos = 0;
    f->locked = 0;
    f->closed = 0;
    f->pos = 0;
    container_init(f->cnt);
    buffer_init(f->buffer);
}

int file_copy(file64* f, char* target) {
    return ERR_OK;
}

int file_fseek(file64* f, long int newpos) {
    f->err = ERR_OK;

    //XXX TODO, in case of status/dir from buffer, rewind/set bufferpointer only */
    //therefore finally move status to file[0xf]->buffer
    //if file is 0xf or is_dir -> nur bufferpointer zurückspulen
    //-> f->bpos = newpos if newpos <= f->bend;
    info_message("setting new pos in file '%s' to '%ld'\n",f->f_glob, newpos);
    if(f->buffered) {
        if(newpos < f->buffer->end) f->buffer->pos = newpos;
    }
    else {
        switch(f->media) {
            case CTYPE_SYS:
                return (f->err = fs_sys_fseek(f, newpos));
            break;
            case CTYPE_D64:
                 return (f->err = fs_d64_fseek(f, newpos));
            break;
        }
        f->err = ERR_UNKNOWN_CTYPE;
    }
    return f->err;
}

/* so rewind is file_fseek(f, file_ftell(f) - num); */
/* iand forward is file_fseek(f, file_ftell(f) + num); */
long int file_ftell(file64* f) {
    return f->buffer->pos;
}

/* expect an already opened and prepared file */
int file_rename(file64* f, char* new_name) {
    f->err = ERR_OK;
    info_message("renaming file '%s' to '%s'\n",f->f_glob, new_name);
    switch(f->media) {
        case CTYPE_SYS:
            f->err = fs_sys_rename(f, f->f_target);
        break;
        case CTYPE_D64:
            f->err = fs_d64_rename(f, f->f_target);
        break;
        default:
            f->err = ERR_UNKNOWN_CTYPE;
        break;
    }
    return f->err;
}

int file_setup_from_dirent(file64* f, dirent64* d) {
    f->err = ERR_OK;
    debug_message("setting up file '%s'\n", f->f_glob);
    dyn_strcpy(&f->petscii_name, d->d_petscii_name);
    dyn_strcpy(&f->ufs_name, d->d_ufs_name);
    f->filetype = d->d_type;
    f->blocks = d->d_blocks;
    f->size = d->d_size;
    f->closed = d->d_closed;
    f->locked = d->d_locked;
    f->detrack = d->d_detrack;
    f->desector = d->d_desector;
    f->desectpos = d->d_desectpos;
    f->track = d->d_track;
    f->sector = d->d_sector;
    return f->err;
}

int file_search(file64* f, uint8_t filter) {
    f->err = ERR_OK;
    info_message("searching file '%s'\n", f->f_glob);
    /* having no filename is bad */
    if(!f->f_glob || !strlen(f->f_glob)) {
        debug_message("no filename given\n");
        f->err = ERR_FILENAME_MISSING;
        return f->err;
    }

    switch(f->media) {
        case CTYPE_SYS:
            f->err = fs_sys_findfirst(f, f->f_glob);
        break;
        case CTYPE_D64:
            f->err = fs_d64_findfirst(f, f->f_glob);
        break;
        default:
            f->err = ERR_UNKNOWN_CTYPE;
        break;
    }
    if(f->cnt->dirsize) {
        f->cnt->dirpos = 0;
        debug_message("we have %d direntries\n", f->cnt->dirsize);
        while(f->cnt->dirpos < f->cnt->dirsize) {
            debug_message("processing direntry #%d\n", f->cnt->dirpos);
            if(filter == FILETYPE_DIR) {
                debug_message("filter is set to dir\n");
                /* only care about dirs */
                switch(f->cnt->dirlist[f->cnt->dirpos]->d_type) {
                    case FILETYPE_DIR:
                        f->err = file_setup_from_dirent(f, f->cnt->dirlist[f->cnt->dirpos]);
                        return f->err;
                    break;
                    default:
                    break;
                }
            }
            else {
                debug_message("filter is set to loadable files\n");
                /* only care about files that contain data */
                switch(f->cnt->dirlist[f->cnt->dirpos]->d_type) {
                    case FILETYPE_SEQ:
                    case FILETYPE_STR:
                    case FILETYPE_PRG:
                        f->err = file_setup_from_dirent(f, f->cnt->dirlist[f->cnt->dirpos]);
                        return f->err;
                    break;
                    default:
                    break;
                }
            }
            f->cnt->dirpos++;
        }
        debug_message("nothing found\n");
    }
    /* get name from f_glob */
    dyn_strcpy(&f->petscii_name, f->f_glob);
    dyn_strcpy(&f->ufs_name, f->f_glob);
    petscii2ascii(f->ufs_name);
    f->err = ERR_FILE_NOT_FOUND;
    return f->err;
}

int file_exists(file64* f) {
    f->err = ERR_OK;
    f->err = file_search(f, 0);
    if(f->cnt->dirlist) f->err = ERR_FILE_EXISTS;
    else f->err = ERR_OK;
    return f->err;
}

int file_available(file64* f) {
    f->err = ERR_OK;
    f->err = file_search(f, 0);
    if(f->cnt->dirlist) f->err = ERR_OK;
    else f->err = ERR_FILE_NOT_FOUND;
    return f->err;
}

/* deletes a file that is proven to exist by a previosu search, so path and ufs/petscii_name are fully qualified and have no more wildcards */
int file_delete(file64* f) {
    f->err = ERR_OK;
    info_message("deleteing file '%s'\n", f->petscii_name);
    if(f->filetype == FILETYPE_DIR) {
         f->err = ERR_FILE_TYPE_MISMATCH;
         return f->err;
    }
    switch(f->media) {
        case CTYPE_SYS:
            f->err = fs_sys_delete(f);
        break;
        case CTYPE_D64:
            f->err = fs_d64_delete(f);
        break;
        default:
            f->err = ERR_UNKNOWN_CTYPE;
        break;
    }
    return f->err;
}

int file_read_block(file64* f, uint8_t *data, long int* num_bytes) {
    f->err = ERR_OK;
    /* read from buffer stream? */
    if(f->buffered) {
    /* read from a file */
    } else {
        switch(f->media) {
            case CTYPE_SYS:
                if(fs_sys_read_block(f, data, num_bytes)) return f->err;
            break;
            case CTYPE_D64:
                if(fs_d64_read_block(f, data, num_bytes)) return f->err;
            break;
            default:
                f->err = ERR_UNKNOWN_CTYPE;
                return f->err;
            break;
        }
    }
    f->pos += *num_bytes;
    return f->err;
}

int file_write_block(file64* f, char *data, long int num_bytes) {
    return f->err;
}

/* either read from file or buff (dirlisting) */
int file_read(file64* f, uint8_t* data) {
    f->err = ERR_OK;
    /* read from buffer stream? */
    if(f->buffered) {
        f->err = buffer_read(f->buffer, data);
    /* read from a file */
    } else {
        switch(f->media) {
            case CTYPE_SYS:
                if(fs_sys_read(f, data)) return f->err;
            break;
            case CTYPE_D64:
                if(fs_d64_read(f, data)) return f->err;
            break;
            default:
                f->err = ERR_UNKNOWN_CTYPE;
                return f->err;
            break;
        }
    }
    f->pos++;
    return f->err;
}

int file_write(file64* f, uint8_t data) {
    f->err = ERR_OK;
    /* write to buffer */
    if(f->buffered) {
         f->err = buffer_write(f->buffer, data, 256);
    /* write to a file */
    } else {
        switch(f->media) {
            case CTYPE_SYS:
                if(fs_sys_write(f, data)) return f->err;
            break;
            case CTYPE_D64:
                if(fs_d64_write(f, data)) return f->err;
            break;
            default:
                f->err = ERR_UNKNOWN_CTYPE;
                return f->err;
            break;
        }
    }
    f->pos++;
    return f->err;
}

int file_open(file64* f) {
    f->err = ERR_OK;

    info_message("opening file '%s'\n",f->f_glob);
    /* file already open? */
    /* should never happen however, as iec_open never lets a fiel stay open */
//    if(f->open) {
//        if(f->f_mode == OPEN_WRITE) {
//            debug_message("file already open\n");
//            f->err = ERR_WRITE_FILE_OPEN;
//            return f->err;
//        }
//        else if (f->f_mode == OPEN_READ) {
//            debug_message("file is already open, closing first.\n");
//            if(file_close(f)) return f->err;
//        }
//    }

    /* handle real files only, no buffered ones (they are always open for read/write) */
    if(!f->buffered) {
        debug_message("file='%s', container_type='%d' fullpath='%s' f_dircmd='%d' buffered='%d'\n",f->f_glob, f->media, f->path, f->f_dircmd, f->buffered);

        /* now search for all files that match f_glob, retrieving a zero dirlist is not an error yet */
        f->err = file_search(f, 0);

        debug_message("ufs_name='%s'  petscii_name='%s'  size='%d'\n",f->ufs_name, f->petscii_name, f->size);
        /* we write to a file? */
        if(f->f_mode == OPEN_WRITE) {
            /* umpf, file exists, as search was okay */
            if(!f->err) {
                /* are we allowed to overwrite? */
                if(f->f_overwrite) {
                    /* file delete checks for right filetype */
                    if((f->err = file_delete(f))) return f->err;
                    /* no files were scratched, should not happen */
                    if(f->err) {
                        f->err = ERR_FILE_EXISTS;
                        return f->err;
                    }
                    /* reset error */
                    f->err = ERR_OK;
                    f->err_track = 0;
                }
                else {
                    f->err = ERR_FILE_EXISTS;
                    return f->err;
                }
            }
            else {
                /* file not found, perfect, so we can create it */
                if(f->err == ERR_FILE_NOT_FOUND) f->err = ERR_OK;
                else return f->err;
            }
            if(f->filetype == FILETYPE_DIR) {
                f->err = ERR_FILE_TYPE_MISMATCH;
                return f->err;
            }
            switch(f->media) {
                case CTYPE_SYS:
                    if(fs_sys_create(f)) return f->err;
                break;
                case CTYPE_D64:
                    if(fs_d64_create(f)) return f->err;
                break;
                default:
                    f->err = ERR_UNKNOWN_CTYPE;
                    return f->err;
                break;
            }
        }
        /* open for read */
        else if(f->f_mode == OPEN_READ) {
            /* no file is bad luck */
            if(f->err) return f->err;
            switch(f->media) {
                case CTYPE_SYS:
                    if(fs_sys_open(f)) return f->err;
                break;
                case CTYPE_D64:
                    if(fs_d64_open(f)) return f->err;
                break;
                default:
                    f->err = ERR_UNKNOWN_CTYPE;
                    return f->err;
                break;
            }
        }
        else {
            fail_message("illegal open-mode '%d'", f->f_mode);
            f->err = ERR_WRITE_FILE_OPEN;
            return f->err;
        }
    }
    debug_message("successful.\n");
    /* set open flag */
    f->open = 1;
    /* reset file pos */
    f->pos = 0;
    return f->err;
}

int file_close(file64* f) {
    f->err = ERR_OK;
    if(f->open) {
        switch(f->media) {
            case CTYPE_SYS:
                if(fs_sys_close(f)) return f->err;
            break;
            case CTYPE_D64:
                if(fs_d64_close(f)) return f->err;
            break;
            default:
                f->err = ERR_UNKNOWN_CTYPE;
                return f->err;
            break;
        }
    }
    f->open = 0;
    debug_message("successful.\n");
    return f->err;
}
