//rename_file
//delete_file
//format

#include "file.h"
#include "fs_d64.h"
#include "debug.h"
#include "container.h"
#include "macros.h"
#include "error.h"
#include "filetypes.h"
#include "dir_funcs.h"
#include "debug_funcs.h"
#include "parse_funcs.h"
#include "dirent_funcs.h"
#include "string_funcs.h"
#include "file_funcs.h"

//int fs_d64_compare(const dirent64** a, const dirent64** b) {
//    return (strcmp((*a)->d_petscii_name, (*b)->d_petscii_name));
//}

static int fs_d64_set_pos(container* cnt, uint8_t track, uint8_t sector) {
    unsigned int offset;
    /* check vals for track and sector */
    cnt->err = ERR_OK;
    if(track > D64_MAX_TRACK || sector > sectors[track]) {
        fail_message("illegal track or sector.\n");
        cnt->err = ERR_ILLEGAL_TRACK_OR_SECTOR_66;
        cnt->err_track = track;
        cnt->err_sector = sector;
        return cnt->err;
    }
    debug_message("track/sector='%d/%d'\n", track, sector);
    /* calculate linear offset in .d64-file */
    offset = track_offsets[track] + (sector * 256);
    debug_message("offset is $%x\n", offset);
    debug_message("filedescriptor is '%d'\n",cnt->file->_fileno);
    if(fseek(cnt->file, offset, SEEK_SET)) {
        fail_message("fssek(): failed\n");
        cnt->err = ERR_READ_ERROR_20;
        return cnt->err;
    }
    debug_message("position set successfully\n");
    return cnt->err;
}

static int fs_d64_read_sector(container* cnt, uint8_t track, uint8_t sector, uint8_t* buf) {
    int err;
    cnt->err = ERR_OK;
    debug_message("reading sector %d/%d\n", track, sector);
    /* set position within .d64 */
    if((cnt->err = fs_d64_set_pos(cnt, track, sector))) {
        fail_message("could not set position\n");
        return cnt->err;
    }
    /* read in 256 bytes */
    err = fread(buf, 1, 256, cnt->file);

    /* did all go smooth? */
    if(err < 256) {
        fail_message("fread(): could only read %d bytes\n", err);
        perror("\n");
        cnt->err = ERR_READ_ERROR_20;
        return cnt->err;
    }
    /* see if sector is partly filled, and calculate sectorsize */
    if(buf[0] == 0) cnt->sectsize = buf[1] + 1;
    else cnt->sectsize = SECTOR_SIZE;
    debug_message("read %d bytes into buf\n", err);
    return cnt->err;
}

static int fs_d64_write_sector(container* cnt, uint8_t track, uint8_t sector, uint8_t* buf) {
    int err;
    cnt->err = ERR_OK;
    debug_message("writing sector %d/%d\n",track,sector);
    /* set position within .d64 */
    if((cnt->err = fs_d64_set_pos(cnt, track, sector))) {
        fail_message("could not set position\n");
        return cnt->err;
    }
    /* write 256 bytes */
    err = fwrite(buf, 1, 256, cnt->file);

    /* did all go smooth? */
    if(err < 256) {
        fail_message("fwrite(): could only write %d bytes\n", err);
        cnt->err = ERR_WRITE_ERROR;
        return cnt->err;
    }
    debug_message("wrote %d bytes from buf\n", err);
    return cnt->err;
}

static int fs_d64_read_bam(container* cnt) {
    debug_message("reading bam...\n");
    cnt->err = fs_d64_read_sector(cnt, D64_BAM_TRACK, D64_BAM_SECTOR, cnt->bam);
    return cnt->err;
}

static int fs_d64_write_bam(container* cnt) {
    debug_message("writing bam...\n");
    cnt->err = fs_d64_write_sector(cnt,D64_BAM_TRACK, D64_BAM_SECTOR, cnt->bam);
    return cnt->err;
}

//XXX auch andere dos typen unterstützen und 40 track/sectors? bam info je nach dostype wo anders -> dostype auflösen und danach entscheiden
static int fs_d64_get_bam_entry(container* cnt, uint8_t track, uint8_t sector) {
    uint8_t byte;
    int offset;
    /* find position within BAM */
    offset = 4 + (track - 1) * 4;
    byte = cnt->bam[offset + (sector >> 3) + 1];
    if(byte & ((uint8_t)1) << (sector & 0x7)) return BAM_FREE;
    return BAM_USED;
}

static int fs_d64_set_bam_entry(container* cnt, uint8_t track, uint8_t sector, int used) {
    uint8_t byte;
    int offset;
    int i;
    int free = 0;

    if(used != BAM_USED && used != BAM_FREE) return 0;

    offset = 4 + (track - 1) * 4;

    byte = cnt->bam[offset + (sector >> 3) + 1];

    byte = byte & (0xff ^ ((uint8_t)1 << (sector & 0x7)));
    byte |= (uint8_t)used << (sector & 0x7);

    cnt->bam[offset + (sector >> 3) + 1] = byte;

    /* count free blocks */
    for(i = 0; i < sectors[track]; i++) {
        if(fs_d64_get_bam_entry(cnt, track, i) == BAM_FREE) free++;
    }
    /* set new value */
    cnt->bam[offset] = free;
    return 1;
}

static int fs_d64_get_free_track_blocks(container* cnt, uint8_t track) {
    int offset;
    if(track < D64_MIN_TRACK || track > D64_MAX_TRACK) return 0;
    offset = 4 + (track - 1) * 4;
    return cnt->bam[offset];
}

int fs_d64_get_free(container *cnt) {
    int track;
    cnt->err = ERR_OK;
    cnt->free = 0;
    if((cnt->err = fs_d64_read_bam(cnt))) return cnt->err;
    for(track = D64_MIN_TRACK; track <= D64_MAX_TRACK; track ++) {
        if(track != D64_DIR_TRACK) cnt->free += fs_d64_get_free_track_blocks(cnt, track);
    }
    return cnt->err;
}

/* fs_d64_allocate_next_block
 * allocates next free block on disk, starting from uint8_t track and uint8_t sector
 * on. The new allocated block will be flagged as used in the BAM */

static int fs_d64_allocate_next_block(container* cnt, uint8_t* track, uint8_t* sector) {
    int free;
    int interleave;
    int i;

    debug_message("allocating next block...\n");

    cnt->err = ERR_OK;

    /* choose interleave */
    if(*track != D64_DIR_TRACK) interleave = FILE_INTERLEAVE;
    else interleave = DIR_INTERLEAVE;
    *sector += interleave;

    if(*sector > sectors[*track]) *sector -= sectors[*track];

    /* let's change track until we have blocks free */
    while(!(free = fs_d64_get_free_track_blocks(cnt, *track))) {
        debug_message("no more free blocks on track %d\n", *track);
        /* seems like dirtrack is full, no more space for more entries */
        if(*track == D64_DIR_TRACK) {
            cnt->err = ERR_DIR_ERROR;
            return cnt->err;
        }
        else {
            (*track)++;
            /* don't touch track 18 */
            if(*track == D64_DIR_TRACK) (*track)++;
            /* reach last track? bad luck! */
            if(*track > D64_MAX_TRACK) {
                cnt->err = ERR_DISK_FULL;
                return cnt->err;
            }
        }
    }
    debug_message("still '%d' free blocks on track %d\n", free, *track);

    /* now as we have a track with free sectors, walk through sector */
    i = sectors[*track];
    *sector = 0;
    while(fs_d64_get_bam_entry(cnt, *track, *sector) == BAM_USED) {
        (*sector)++;
        if(*sector > sectors[*track]) *sector -= sectors[*track];
        i--;
        if(i == 0) {
            /* OOOPS, something really weired happend, bam must be corrupted */
            fail_message("can't find free block, though should be free (bam currupted?)\n");
            cnt->err = ERR_DIR_ERROR;
            return cnt->err;
        }
    }
    /* mark found block as used */
    fs_d64_set_bam_entry(cnt, *track, *sector, BAM_USED);
    debug_message("next free block is @%d/%d, interleave was %d\n", *track, *sector, interleave);
    return cnt->err;
}

/*
static int fs_d64_display_bam(container* cnt) {
    uint8_t track;
    int i;
    int val;
    for(track = D64_MIN_TRACK; track <= D64_MAX_TRACK; track++) {
        for(i = 0; i < sectors[track]; i++) {
            val = fs_d64_get_bam_entry(cnt, track, i);
            if(val == BAM_FREE) printf(" ");
            else printf(".");
        }
        printf("\n");
    }
    return 1;
}
*/

int fs_d64_fseek(file64* f, long int newpos) {
    container* cnt = f->cnt;

    f->err = ERR_OK;

    /* set startposition */
    cnt->track = f->track;
    cnt->sector = f->sector;

    f->pos = 0;

    while(!f->err) {
        /* get sector */
        if((f->err = fs_d64_read_sector(cnt, cnt->track, cnt->sector, cnt->sectbuf))) {
            break;
        }

        /* fetch ts-link */
        cnt->track   = cnt->sectbuf[0];
        cnt->sector  = cnt->sectbuf[1];

        /* skip whole sector? */
        if(newpos - f->pos >= (cnt->sectsize - cnt->sectpos)) {
            /* reset sectpos */
            cnt->sectpos = 2;
            f->pos += (cnt->sectsize - cnt->sectpos);
        } else {
            cnt->sectpos = (newpos - f->pos) + 2;
            f->pos = newpos;
            break;
        }
    }
    return f->err;
}

int fs_d64_read_block(file64* f, char* data, long int* num_bytes) {
    return ERR_OK;
}

int fs_d64_write_block(file64* f, char* data, long int* num_bytes) {
    return ERR_OK;
}

/* fs_d64_read
 * will read the next byte from the currently open file desribed by file64* f into *data.
 * The function will follow the sectorchain until finished and advance cnt->track and cnt->sector
 * when necessary */

int fs_d64_read(file64* f, char* data) {
    f->err = ERR_OK;
    container* cnt = f->cnt;

    /* read byte first */
    *data = cnt->sectbuf[cnt->sectpos];
    cnt->sectpos++;

    debug_message("sectpos: %d   sectsize: %d\n",cnt->sectpos,cnt->sectsize);

    if(cnt->sectpos == cnt->sectsize) {
        debug_message("reached end of sector\n");
        /* so try to move to next sector */
        /* fetch ts-link */
        cnt->track   = cnt->sectbuf[0];
        cnt->sector  = cnt->sectbuf[1];
        /* oops, no next sector */
        if(!cnt->track) {
            debug_message("sector-chain finshed\n");
            cnt->sectpos--;
            f->err = ERR_EOF;
            return f->err;
        }
        /* follow sector chain */
        if((f->err = fs_d64_read_sector(cnt, cnt->track, cnt->sector, cnt->sectbuf))) {
            return f->err;
        }
        cnt->sectpos = 2;
    }

    return f->err;
}

/* fs_d64_write
 * will write data to the currently open filedescribed by file64*. If one block is filled
 * it will allocate a new one until no more blocks are available. Thus track, sector, track_link
 * and sector_link are touched. */

int fs_d64_write(file64* f, char data) {
    f->err = ERR_OK;
    container* cnt = f->cnt;

    /* no more space in buffer? */
    if(cnt->sectpos == cnt->sectsize) {
        /* allocate a new block, therefore set up ts-link */
        cnt->track_link = cnt->track;
        cnt->sector_link = cnt->sector;
        if((f->err = fs_d64_allocate_next_block(cnt, &cnt->track_link, &cnt->sector_link))) {
            return f->err;
        }
        /* count blocks up */
        f->blocks++;

        /* set ts-link in finished block */
        cnt->sectbuf[0] = cnt->track_link;
        cnt->sectbuf[1] = cnt->sector_link;

        /* write finished sector to disc */
        if((f->err = fs_d64_write_sector(cnt, cnt->track, cnt->sector, cnt->sectbuf))) {
            return f->err;
        }

        /* start with an empty block */
        memset(cnt->sectbuf, 0, 256);

        /* set up new position */
        cnt->track    = cnt->track_link;
        cnt->sector   = cnt->sector_link;
        cnt->sectpos  = 2;
        cnt->sectsize = SECTOR_SIZE;
    }

    cnt->sectbuf[cnt->sectpos] = data;
    cnt->sectpos++;

    return f->err;
}

static int fs_d64_update_dirent(container* cnt, file64* f) {
    int err = ERR_OK;

    /* read in sector the dirent is on */
    if((err = fs_d64_read_sector(cnt, f->detrack, f->desector, cnt->sectbuf))) return err;
    /* now write dirent-content  to sectbuf */
    cnt->sectbuf[f->desectpos + D64_DTRACK]  = f->track;
    cnt->sectbuf[f->desectpos + D64_DSECTOR] = f->sector;
    cnt->sectbuf[f->desectpos + D64_DTYPE]   = f->filetype;
    if(f->closed) {
        cnt->sectbuf[f->desectpos + D64_DTYPE] = cnt->sectbuf[f->desectpos + D64_DTYPE] | 0x80;
    }
    /* blank filename */
    memset(&cnt->sectbuf[f->desectpos + D64_DNAME], 0xa0, 0x10);
    /* add filename */
    memcpy(&cnt->sectbuf[f->desectpos + D64_DNAME], f->petscii_name, strlen(f->petscii_name));
    /* add size information */
    cnt->sectbuf[f->desectpos + D64_DBLOCKS + 1] = f->blocks / 256;
    cnt->sectbuf[f->desectpos + D64_DBLOCKS + 0] = f->blocks & 0xff;

    debug_print_hex(&cnt->sectbuf[f->desectpos], 32, "DIR:", 32);

    /* update size info if it is the last dir-sector */
    if(cnt->sectbuf[0] == 0) cnt->sectbuf[1] = SECTOR_SIZE - 1;

    /* write back to disc */
    if((err = fs_d64_write_sector(cnt, f->detrack, f->desector, cnt->sectbuf))) return err;
    debug_message("updated dirent data: c64_name='%s', startpos='%d/%d', closed='%d', filetype='%s', blocks='%d'\n",
                  f->petscii_name, f->track, f->sector, f->closed, filetypes[f->filetype], f->blocks);

    memset(cnt->sectbuf, 0, 256);
    return err;
}

static int fs_d64_readdir(container* cnt, file64* f) {
    int a;
    int err;
    /* still data to process on current sector? */
    if(cnt->sectpos >= cnt->sectsize) {
        /* we have a next ts-link? */
        if(cnt->track_link) {
            /* follow ts-link */
            cnt->track  = cnt->track_link;
            cnt->sector = cnt->sector_link;
            debug_message("trying to read new sector %d/%d\n",cnt->track,cnt->sector);
            /* get next sector */
            if((err = fs_d64_read_sector(cnt, cnt->track, cnt->sector, cnt->sectbuf))) {
                return err;
            }
            /* first, get ts-link from new sector */
            cnt->track_link  = cnt->sectbuf[0];
            cnt->sector_link = cnt->sectbuf[1];

            /* ignore size set by read_sector() always walk through whole block, as we have a dir and size does not matter here */
            cnt->sectsize = SECTOR_SIZE;
            /* start at pos 0, to have 8 blocks with $20 bytes size, first two bytes (tslink on first entry) will be ignored anyway, but counting is easier */
            cnt->sectpos  = 0;
        }
        else {
            debug_message("sector-chain finshed\n");
            return ERR_EOF;
        }
    }

    debug_message("sectpos='%d' sectsize='%d'\n", cnt->sectpos, cnt->sectsize);

    /* populate direntry with data */
    f->desectpos = cnt->sectpos;
    f->detrack   = cnt->track;
    f->desector  = cnt->sector;
    f->filetype  = cnt->sectbuf[cnt->sectpos + D64_DTYPE] & 0x7;
    f->track     = cnt->sectbuf[cnt->sectpos + D64_DTRACK];
    f->sector    = cnt->sectbuf[cnt->sectpos + D64_DSECTOR];
    f->blocks    = cnt->sectbuf[cnt->sectpos + D64_DBLOCKS + 1] * 256 + cnt->sectbuf[cnt->sectpos + D64_DBLOCKS];
    f->size      = f->blocks * (SECTOR_SIZE - 2);
    if( (cnt->sectbuf[cnt->sectpos + D64_DTYPE] & 0x80)) f->closed = 1;
    else f->closed = 0;
    if( (cnt->sectbuf[cnt->sectpos + D64_DTYPE] & 0x40)) f->locked = 1;
    else f->locked = 0;
    /* copy filename */
    for(a = 0; a < MAX_C64_NAME; a++) {
        if(cnt->sectbuf[cnt->sectpos + D64_DNAME + a] == 0xa0) break;
    }
    dyn_strncpy(&f->petscii_name, (char*)&cnt->sectbuf[cnt->sectpos + D64_DNAME], a);

    /* advance to next direntry on sector */
    cnt->sectpos += 32;
    return ERR_OK;
}

static int fs_d64_create_dirent(container* cnt, file64* f) {
    /* set up things */
    cnt->track       = D64_DIR_TRACK;
    cnt->track_link  = D64_DIR_TRACK;
    cnt->sector      = D64_DIR_SECTOR;
    cnt->sector_link = D64_DIR_SECTOR;
    cnt->sectpos     = 0;
    cnt->sectsize    = SECTOR_SIZE;

    if((f->err = fs_d64_read_bam(cnt))) return f->err;
    /* walk through dir to find reusable direntries, position for a found entry will be set appropriately */
    while(!(f->err = fs_d64_readdir(cnt, f))) {
        if(!f->filetype) break;
    }
    /* nothing free? */
    if(f->err == ERR_EOF) {
        debug_message("last block of dir reached, need to extend dir from t/s='%d/%d' on\n", cnt->track, cnt->sector);
        /* extend dir with new block */
        cnt->track_link  = cnt->track;
        cnt->sector_link = cnt->sector;
        if((f->err = fs_d64_allocate_next_block(cnt, &cnt->track_link, &cnt->sector_link))) {
            return f->err;
        }
        /* update ts-link */
        cnt->sectbuf[0] = cnt->track_link;
        cnt->sectbuf[1] = cnt->sector_link;

        /* write sector back to update ts-link */
        fs_d64_write_sector(cnt, cnt->track, cnt->sector, cnt->sectbuf);

        /* set position for new direntry */
        f->detrack   = cnt->track_link;
        f->desector  = cnt->sector_link;
        f->desectpos = 0;
    }
    debug_message("free dirent found at t/s='%d/%d' @$%02x\n", f->detrack, f->desector, f->desectpos);

    /* allocate first free block on disk */
    cnt->track_link  = 0;
    cnt->sector_link = 0;
    fs_d64_allocate_next_block(cnt, &cnt->track_link, &cnt->sector_link);

    /* ...and create it on disc */
    fs_d64_update_dirent(cnt, f);

    /* count first allocated block */
    f->blocks++;
    return ERR_OK;
}

/* fs_d64_delete
 * deletes a single file described with file64* f with a full name (no wildcards) usually found by file_search. */

int fs_d64_delete(file64* f) {
    container* cnt = f->cnt;
    f->err = ERR_OK;
    if((f->err = fs_d64_read_bam(cnt))) return f->err;
    /* get first sector of direntry */
    cnt->track  = f->track;
    cnt->sector = f->sector;
    debug_message("walking through sectorchain...\n");
    /* read sector */
    while(!(f->err = fs_d64_read_sector(cnt, cnt->track, cnt->sector, cnt->sectbuf))) {
        /* save link to next sector */
        cnt->track_link  = cnt->sectbuf[0];
        cnt->sector_link = cnt->sectbuf[1];
        debug_message("blanking sector %d/%d...\n", cnt->track, cnt->sector);
        /* clear sector */
        memset(cnt->sectbuf, 0, 256);
        /* write back sector */
        if((f->err = fs_d64_write_sector(cnt, cnt->track, cnt->sector, cnt->sectbuf))) return f->err;
        /* mark block as free in BAM */
        if(!fs_d64_set_bam_entry(cnt, cnt->track, cnt->sector, BAM_FREE)) {
            f->err = ERR_DIR_ERROR;
            return f->err;
        }
        /* now follow to new ts-link */
        cnt->track  = cnt->track_link;
        cnt->sector = cnt->sector_link;
        if(!cnt->track) break;
    }
    debug_message("sector-chain finshed\n");
    if(f->err && f->err != ERR_EOF) return f->err;
    debug_message("updating bam\n");
    if((f->err = fs_d64_write_bam(cnt))) return f->err;

    /* 'wipe' dirent */
    f->filetype = 0;
    f->closed = 0;
    debug_message("updating dir\n");
    fs_d64_update_dirent(cnt, f);
    return f->err;
}

static int fs_d64_set_header(container* cnt, char* header, char* id) {
    int i;
    int err;
    if((err = fs_d64_read_bam(cnt))) return err;
    memset(&cnt->bam[D64_BAM_HEADER], 0xa0, MAX_C64_NAME);
    i = strlen(header);
    if(i > MAX_C64_NAME) i = MAX_C64_NAME;
    memcpy(&cnt->bam[D64_BAM_HEADER], header, i);
    i = strlen(id);
    if(i > 5) i = 5;
    memcpy(&cnt->bam[D64_BAM_ID], id, i);
    if((err = fs_d64_write_bam(cnt))) return err;
    return ERR_OK;
}

int fs_d64_get_header(container *cnt, char* path) {
    int err;
    int i;
    if((err = fs_d64_read_bam(cnt))) return err;
    cnt->header = realloc(cnt->header, (MAX_C64_NAME + 1) * sizeof(char));
    memcpy(cnt->header, &(cnt->bam)[D64_BAM_HEADER], MAX_C64_NAME);
    cnt->header[MAX_C64_NAME] = '\0';
    for(i = 0; i < 5; i++) {
        cnt->id[i] = cnt->bam[D64_BAM_ID + i] & 0x7f;
    }
    cnt->id[5] = '\0';
    debug_message("header information extracted header='%s' id='%s'\n", cnt->header, cnt->id);
    return ERR_OK;
}

int fs_d64_findfirst(file64* f, char* f_glob) {
    int i = 0;
    container* cnt = f->cnt;
    /* a file objected is needed for readdir, so we create a new one to don't touch our own file */
    dirent64* d = dirent_create();
    file64* fd = file_create();
    //set startpos
    cnt->track_link  = D64_DIR_TRACK;
    cnt->sector_link = D64_DIR_SECTOR;
    cnt->track       = D64_DIR_TRACK;
    cnt->sector      = D64_DIR_SECTOR;
    //no data read yet
    cnt->sectpos     = 0;
    cnt->sectsize    = SECTOR_SIZE;

    file_init(fd);

    debug_message("opened dir on diskimage '%s', searchmask is '%s'\n",f->path, f_glob);
    while(!(f->err = fs_d64_readdir(cnt, fd))) {
        debug_message("have direntry: '%s' size='%d bytes (%d blocks)' locked='%d' closed='%d' filetype='%s' sectorchain_start='%d/%d, pos_in_dir='%d/%d @$%02x'\n",
                          fd->petscii_name, fd->size, fd->blocks, fd->locked, fd->closed, filetypes[fd->filetype], fd->track, fd->sector, fd->detrack, fd->desector, fd->desectpos);
        dirent_setup_from_file(d, fd);
        /* type needs to be set (else it is a deleted file) and name must match */
        if(d->d_type && file_select(d->d_petscii_name, f_glob)) {
            cnt->dirlist = (struct dirent64**)realloc((void*)(cnt->dirlist), (size_t)((i + 1) * sizeof(struct dirent64*)));
            if(cnt->dirlist == NULL) {
                f->err = ERR_DRIVE_NOT_READY;
                break;
            }
            (cnt->dirlist)[i] = dirent_create();
            /* copy content of dirent */
            *(cnt->dirlist)[i] = *d;
            if((cnt->dirlist)[i] == NULL) {
                f->err = ERR_DRIVE_NOT_READY;
                break;
            }
            debug_message("MATCHING direntry: '%s' size='%d bytes (%d blocks)' locked='%d' closed='%d' filetype='%s' sectorchain_start='%d/%d, pos_in_dir='%d/%d @$%02x'\n",
                          d->d_petscii_name, d->d_size, d->d_blocks, d->d_locked, d->d_closed, filetypes[d->d_type], d->d_track, d->d_sector, d->d_detrack, d->d_desector, d->d_desectpos);
            i++;
        }
    }
    dirent_destruct(d);
    file_destruct(fd);
    if(f->err && f->err != ERR_EOF) return f->err;
    f->err = ERR_OK;
    cnt->dirsize = i;
    debug_message("found %d entries'\n", cnt->dirsize);
    //qsort((void*)(dir->dirlist),(size_t)i,sizeof(struct dirent64*),(void*)file_compare);
    return f->err;
}

int fs_d64_close(file64* f) {
    container* cnt = f->cnt;
    f->err = ERR_OK;
    if(!f->open) {
        f->err = ERR_FILE_NOT_OPEN;
        return f->err;
    }
    if(!f->f_dircmd) {
        /* write back bam and dirent with final blocksize in time */
        if(f->f_mode == OPEN_WRITE) {
            /* set last ts-link */
            cnt->sectbuf[0] = 0x00;
            cnt->sectbuf[1] = cnt->sectpos - 1;
            /* write last sector */
            fs_d64_write_sector(cnt, cnt->track, cnt->sector, cnt->sectbuf);
            /* blank buffer */
            memset(cnt->sectbuf, 0, 256);
            /* write changed bam */
            fs_d64_write_bam(cnt);
            /* set closed flag */
            f->closed = 1;
            /* update dirent */
            fs_d64_update_dirent(cnt, f);
        }
    }
    return f->err;
}

int fs_d64_format(container* cnt, char* header, char* id) {
    int err = ERR_OK;
    int track, sector;
    if(!err) err = fs_d64_set_header(cnt, header, id);
    err = fs_d64_read_bam(cnt);
    /* clear bam */
    for(track = D64_MIN_TRACK; track <= D64_MAX_TRACK; track++) {
        for(sector = 0; sector < sectors[track]; sector++) {
            fs_d64_set_bam_entry(cnt, track, sector, BAM_FREE);
        }
    }
    /* allocate bam block */
    fs_d64_set_bam_entry(cnt, D64_DIR_TRACK, 0, BAM_USED);
    /* allocate dir block */
    fs_d64_set_bam_entry(cnt, D64_DIR_TRACK, 1, BAM_USED);

    err = fs_d64_write_bam(cnt);
    return err;
}

int fs_d64_rename(file64* f, char* newname) {
    container *cnt = f->cnt;
    f->err = ERR_OK;
    dyn_strncpy(&f->petscii_name, newname, MAX_C64_NAME);
    f->err = fs_d64_update_dirent(cnt, f);
    return f->err;
}

/* opens a file on a .d64. We expect file64 containing already a valid ifile64 from a file_search */
int fs_d64_open(file64* f) {
    container* cnt = f->cnt;
    f->err = ERR_OK;
    /* set startposition */
    cnt->track    = f->track;
    cnt->sector   = f->sector;
    /* read in first block */
    f->err = fs_d64_read_sector(cnt, cnt->track, cnt->sector, cnt->sectbuf);
    cnt->sectpos  = 2;
    return f->err;
}

int fs_d64_create(file64* f) {
    f->err = ERR_OK;
    container* cnt = f->cnt;
    /* create dirent and allocate first block */
    if((f->err = fs_d64_create_dirent(cnt, f))) return f->err;
    memset(cnt->sectbuf, 0, 256);
    /* init buffer positions */
    cnt->sectpos  = 2;
    cnt->sectsize = SECTOR_SIZE;
    /* set startposition */
    cnt->track    = f->track;
    cnt->sector   = f->sector;
    return f->err;
}

