#include "iec_funcs.h"
#include "file_funcs.h"
#include "network_funcs.h"
#include "iec.h"
#include "file.h"
#include "debug_funcs.h"
#include "buffer_funcs.h"
#include "parse_funcs.h"
#include "dos_funcs.h"
#include "path_funcs.h"

static char* iec_sa_name(int sa) {
    switch(sa) {
        case IEC_DATA:
            return "DATA";
        break;
        case IEC_OPEN:
            return "OPEN";
        break;
        case IEC_CLOSE:
            return "CLOSE";
        break;
        case 0:
            return "UNSET";
        break;
        default:
            return "UNKNOWN";
        break;
    }
}

static char* iec_job_name(int job) {
    switch(job) {
        case IEC_IDLE:
            return "IDLE";
        break;
        case IEC_LISTEN:
            return "LISTEN";
        break;
        case IEC_TALK:
            return "TALK";
        break;
        case IEC_UNTALK:
            return "UNTALK";
        break;
        case IEC_UNLISTEN:
            return "UNLISTEN";
        break;
        default:
            return "UNKNOWN";
        break;
    }
}

static void iec_error(drive* c, uint8_t err) {
    /* set up a error packet */
    debug_message("\n");
    c->out.size = 1;
    c->out.type = PACKET_ERROR;
    c->out.err = err;
}

static void iec_data(drive* c, int size, int err) {
    /* set up a data packet */
    debug_message("\n");
    c->out.size = size;
    c->out.type = PACKET_DATA;
    c->out.err = err;
}

static void iec_close(drive* c) {
    debug_message("\n");
    if(c->files[c->lfn]) {
        /* close the file */
        dos_close(c, c->files[c->lfn]);
        /* end benchmarking */
        debug_end_measure(c->files[c->lfn]);
        /* before we destruct the file */
        file_destruct(c->files[c->lfn]);
        c->files[c->lfn] = NULL;
    }
    else {
        debug_message("nothing to close, no fileobject\n");
    }
}

static void iec_init(drive* c) {
    debug_message("\n");
    /* init filespec */
    buffer_init(c->filespec);
}

static void iec_open(drive* c) {
    debug_message("\n");
    /* someone left the logical file open? */
    /* XXX TODO, actually we should issue a file open error here? */
    /* if we really want to close a file, we use the close command or I: */
    if(c->files[c->lfn] && c->files[c->lfn]->open) iec_close(c);
    /* create new file object */
    c->files[c->lfn] = file_create();
    /* init it */
    file_init(c->files[c->lfn]);
    /* and start benchmarking */
    debug_start_measure(c->files[c->lfn]);
    /* now let the dos open the file */
    dos_open(c, c->files[c->lfn], c->lfn);
}

static void iec_handle_command(drive *c) {
    uint8_t new_job = c->in.command & 0xf0;
    uint8_t lfn = c->in.command & 0x0f;

    debug_message("\n");

    /* always react on new listen/talk and drop old state */
    if(new_job == IEC_LISTEN || new_job == IEC_TALK) {
        c->job = new_job;
        c->SA = 0;
        iec_error(c, IEC_OK);
    }
    else if(c->job == IEC_LISTEN) {
        debug_message("still listening\n");
        switch(new_job) {
            case IEC_UNLISTEN:
                /* open dos command or file */
                if(c->SA == IEC_OPEN) {
                    iec_open(c);
                }
                iec_error(c, IEC_OK);
                c->SA = 0;
                c->job = IEC_IDLE;
            break;
            case IEC_DATA:
            case IEC_OPEN:
            case IEC_CLOSE:
                c->lfn = lfn;
                c->SA = new_job;
                if(new_job == IEC_OPEN) {
                    iec_init(c);
                }
                /* close any file# */
                if(new_job == IEC_CLOSE) {
                    iec_close(c);
                }
                iec_error(c, IEC_OK);
            break;
        }
    }
    else if(c->job == IEC_TALK) {
        debug_message("still talking\n");
        switch(new_job) {
            case IEC_DATA:
                c->lfn = lfn;
                c->SA = new_job;
                iec_error(c, IEC_OK);
            break;
            case IEC_UNTALK:
                c->SA = 0;
                c->job = IEC_IDLE;
                iec_error(c, IEC_OK);
            break;
        }
    }
    /* we can accept a new job */
    else if(c->job == IEC_IDLE) {
        /* but only allow nice jobs */
        if(new_job == IEC_UNLISTEN || new_job == IEC_UNTALK) {
            iec_error(c, IEC_OK);
        }
    }
    else {
        debug_message("illegal command: job: %d\n",new_job);
    }
    debug_message("job: %s ($%02x)  SA: %s ($%02x)  lfn: $%02x\n", iec_job_name(c->job), c->job, iec_sa_name(c->SA), c->SA, c->lfn);
}

static void iec_handle_request(drive *c) {
    int i = 0;
    int err = ERR_OK;
    file64* f = c->files[c->lfn];
    debug_message("job: %s ($%02x)  SA: %s ($%02x)  lfn: $%02x\n", iec_job_name(c->job), c->job, iec_sa_name(c->SA), c->SA, c->lfn);
    if(c->job == IEC_TALK) {
        if(c->SA == IEC_DATA) {
            debug_message("talking...\n");
            if(c->lfn == IEC_ERROR_CHANNEL) {
                /* send status */
                for(i = 0; i < c->in.req_size; i++) {
                    if((err = buffer_read(c->status, &c->out.data[i]))) {
                        /* count last byte that came under error */
                        i++;
                        /* reset status after no bytes were read anymore from status */
                        dos_set_status(c, ERR_OK, 0, 0);
                        break;
                    }
                }
            }
            else {
                if(f && !f->err) {
                    /* send requested data */
                    for(i = 0; i < c->in.req_size; i++) {
                        if((err = file_read(f, &c->out.data[i]))) {
                            /* count last byte that came under error */
                            i++;
                            break;
                        }
                    }
                }
            }
        }
        /* error handling */
        if(f && f->err == ERR_EOF) {
            f->err = ERR_OK;
            iec_data(c, i, IEC_EOF);
        }
        else if(i) {
            iec_data(c, i, IEC_OK);
        }
        else {
            iec_data(c, 1, IEC_TIMEOUT | IEC_READ);
        }
    }
}

static void iec_handle_data(drive *c) {
    int i = 0;
    file64* f = c->files[c->lfn];
    debug_message("job: %s ($%02x)  SA: %s ($%02x)  lfn: $%02x\n", iec_job_name(c->job), c->job, iec_sa_name(c->SA), c->SA, c->lfn);
    if(c->job == IEC_LISTEN) {
        switch(c->SA) {
            case IEC_DATA:
                if(c->lfn == IEC_ERROR_CHANNEL) {
                    debug_message("listening for command...\n");
                    /* receive a command or filespec */
                    for(i = 0; i < c->in.size; i++) {
                        buffer_write(c->filespec, c->in.data[i], 32);
                    }
                    iec_error(c, IEC_OK);
                }
                else {
                    debug_message("listening for data...\n");
                    if(f && !f->err) {
                        /* receive data */
                        for(i = 0; i < c->in.size; i++) {
                            file_write(f, c->in.data[i]);
                        }
                    }
                    /* error handling */
                    if(f && !f->err) {
                        iec_error(c, IEC_OK);
                    } else {
                        iec_error(c, IEC_WRITE | IEC_TIMEOUT);
                    }
                }
            break;
            case IEC_OPEN:
                debug_message("listening for filespec...\n");
                /* receive a command or filespec */
                for(i = 0; i < c->in.size; i++) {
                    buffer_write(c->filespec, c->in.data[i], 32);
                }
                iec_error(c, IEC_OK);
            break;
        }
    }
    if(c->job == IEC_TALK) {
        if(c->SA == IEC_DATA) {
            /* actually e should never receive data when talking, but in the very case of having still
             * data to receive and starting a command, the c64 will flush and send the remeining data
             * back, Thus we are able to react on a interrupted transfer and in case rewind our
             * file/bufferstream (get#1-issue that interrupts after each byte with an untalk/talk
             * sequence). Yet however do nothing but send an ack to make c64 happy and not hang. */
            debug_message("talking...\n");
            if(c->lfn == IEC_ERROR_CHANNEL) {
            }
            else {
            }
        }
        iec_error(c, IEC_OK);
    }
}

void iec_process_packet(drive* c) {
    /* clear outgoing packet */
    c->out.size = 0;
    c->out.type = PACKET_NONE;

    switch(c->in.type) {
        case PACKET_COMMAND:
            /* things that are send under ATN */
            iec_handle_command(c);
        break;
        case PACKET_REQUEST:
            /* data requests */
            iec_handle_request(c);
        break;
        case PACKET_DATA:
            /* data we receive */
            iec_handle_data(c);
        break;
    }
}
